#	vim:fileencoding=utf-8
# (c) 2013 Michał Górny <mgorny@gentoo.org>
# Released under the terms of the 2-clause BSD license.

import os, os.path, re, tempfile

class Whitespace(str):
	def __init__(self, s):
		self.removed = False
		str.__init__(self)

	def __repr__(self):
		return 'Whitespace(%s)' % str.__repr__(self)


class Value(object):
	def __init__(self, f_name, l_name=None):
		self.full_name = f_name
		self.local_name = l_name or f_name
		self.removed = False

	def __repr__(self):
		return 'Value(full_name=%s, local_name=%s, removed=%s)' % (
			self.full_name,
			self.local_name,
			self.removed,
		)

	def __str__(self):
		assert(not self.removed)
		return self.local_name


def get_previous_val_index(l, v):
	"""
	Return index of value in list l that lexically precedes v,
	or -1 if no value precedes it.

	>>> get_previous_val_index([Value('a'), Value('b')], Value('c'))
	1
	>>> get_previous_val_index([Value('a'), Value('c')], Value('b'))
	0
	>>> get_previous_val_index([Value('c'), Value('a')], Value('b'))
	1
	>>> get_previous_val_index([Value('c'), Value('b')], Value('a'))
	-1
	"""
	def sort_key(x):
		if isinstance(x, Whitespace):
			# sort whitespace out to the end
			return 'Z'
		elif isinstance(x, Value):
			return x.local_name
		else:
			return x.local_prefix

	sorted_values = sorted(l + [v], key=sort_key)
	idx = sorted_values.index(v)
	if idx == 0:
		return -1
	else:
		return l.index(sorted_values[idx-1])


class Group(object):
	def __init__(self, f_prefix, l_prefix, values):
		self.full_prefix = f_prefix
		self.local_prefix = l_prefix
		self.values = values

	def add_sorted(self, v):
		self.values.insert(get_previous_val_index(self.values, v)+1, v)

	@property
	def removed(self):
		return all([x.removed for x in self.values])

	def __iter__(self):
		for x in self.values:
			yield x

	def __repr__(self):
		return 'Group(full_prefix=%s, local_prefix=%s, values=%s)' % (
			self.full_prefix,
			self.local_prefix,
			self.values,
		)

	def __str__(self):
		vals = [str(x) for x in self.values if not x.removed]

		if len(vals) > 1:
			vals = '{%s}' % ','.join(vals)
		else:
			vals = vals[0]

		return ''.join((self.local_prefix, vals))


class PythonCompat(object):
	def __init__(self):
		self.nodes = []

	def append(self, n):
		self.nodes.append(n)

	def add(self, impl_name):
		# first, try adding to an existing group
		# longer groups come first, so that should be good enough
		for g in self.groups:
			if impl_name.startswith(g.full_prefix):
				g.add_sorted(Value(impl_name, impl_name[len(g.full_prefix):]))
				return

		# then, try splitting something else
		for v in reversed(sorted(self, key=lambda x: len(x.full_name))):
			cpfx = os.path.commonprefix((impl_name, v.full_name))
			# only those with common prefix
			if not cpfx:
				continue
			# only split in global scope, don't nest
			if v not in self.nodes:
				continue
			# don't split mid-version if maintainer didn't do that already
			mid_ver_groups = [x for x in self.groups
				if x.local_prefix.endswith('_')]
			if cpfx[-1] == '_' and not any(mid_ver_groups):
				continue

			suff1 = v.full_name[len(cpfx):]
			suff2 = impl_name[len(cpfx):]
			# don't split in middle of name (e.g. py{py,thon})
			if not suff1[0].isdigit() or not suff2[0].isdigit():
				continue

			# don't create the group if the maintainer could have created
			# one already but didn't (so he likely doesn't want that)
			# e.g. if he has 'python2_5 python2_6', don't do '{2_6,2_7}'
			group_candidates = [x for x in self.nodes
				if isinstance(x, Value) and x.full_name.startswith(cpfx)]
			if len(group_candidates) > 1:
				continue

			v1 = Value(v.full_name, suff1)
			v2 = Value(impl_name, suff2)

			i = self.nodes.index(v)
			self.nodes[i] = Group(cpfx, cpfx,
					sorted([v1, v2], key=lambda x: x.local_name))
			return

		# add it (sorted!)
		v = Value(impl_name, impl_name)
		i = get_previous_val_index(self.nodes, v)

		self.nodes.insert(i+1, v)
		self.nodes.insert(i+1 if i != -1 else 1, Whitespace(' '))

	def remove(self, impl_name):
		for i in self:
			if i.full_name == impl_name:
				i.removed = True

	@property
	def groups(self):
		def subiter(g):
			for x in g:
				if isinstance(x, Group):
					for y in subiter(x):
						yield y
					yield x

		return subiter(self.nodes)

	def __iter__(self):
		def subiter(g):
			for x in g:
				if isinstance(x, Group):
					for y in subiter(x):
						if not y.removed:
							yield y
				elif isinstance(x, Value):
					if not x.removed:
						yield x

		return subiter(self.nodes)

	def __repr__(self):
		return repr(self.nodes)

	def __str__(self):
		first = True
		for i, x in enumerate(self.nodes):
			if not isinstance(x, Whitespace):
				if x.removed:
					if not first:
						self.nodes[i-1].removed = True
					else:
						self.nodes[i+1].removed = True
				first = False

		return ''.join([str(x) for x in self.nodes if not x.removed])


def parse_item(s):
	depth = 0
	curr = ['']
	items = []
	values = [[]]
	out = []
	had_text = []

	def commit_value():
		if had_text:
			values[-1].append(Value(''.join(curr), curr[-1]))
			had_text.pop()

	for c in s:
		if c == '{':
			depth += 1
			curr.append('')
			values.append([])
		elif c == '}':
			if depth == 0:
				raise ValueError("Unmatched closing brace '}'")

			commit_value()
			if values[-1]:
				values[-2].append(Group(
					''.join(curr[:-1]),
					curr[-2],
					values[-1],
				))

			depth -= 1
			curr.pop()
			values.pop()
		elif c == ',':
			if depth == 0:
				raise ValueError("Comma ',' outside brace")
			commit_value()
			curr[-1] = ''
		else:
			had_text = [True]
			curr[-1] += c

	if depth != 0:
		raise ValueError("Unmatched opening brace '{'")
	commit_value()

	return values[0][0]


ws_split_re = re.compile(r'(\s+)')

def parse(s):
	out = PythonCompat()

	data = ws_split_re.split(s)
	for i, w in enumerate(data):
		if not w:
			pass
		elif i % 2:
			# 1, 3, 5 are whitespace
			out.append(Whitespace(w))
		else:
			# 0, 2, 4 are the real deal
			out.append(parse_item(w))

	return out


def add_impl(s, new):
	"""
	>>> add_impl('pypy1_9', 'python3_3')
	'pypy1_9 python3_3'
	>>> add_impl('python2_7', 'pypy2_0')
	'pypy2_0 python2_7'
	>>> add_impl('python2_6 python2_7 python3_2 pypy1_9', 'python3_3')
	'python2_6 python2_7 python3_2 python3_3 pypy1_9'
	>>> add_impl('python2_6 python2_7 python3_4 pypy1_9', 'python3_3')
	'python2_6 python2_7 python3_3 python3_4 pypy1_9'
	>>> add_impl('python2_{6,7} python3_{1,2}', 'python3_3')
	'python2_{6,7} python3_{1,2,3}'
	>>> add_impl('python2_{6,7} python3_2', 'python3_3')
	'python2_{6,7} python3_{2,3}'
	>>> add_impl('python2_{6,7} python3_4', 'python3_3')
	'python2_{6,7} python3_{3,4}'
	>>> add_impl('python{2_6,2_7,3_2} pypy1_9', 'python3_3')
	'python{2_6,2_7,3_2,3_3} pypy1_9'
	>>> add_impl('python{2_6,2_7,3_4} pypy1_9', 'python3_3')
	'python{2_6,2_7,3_3,3_4} pypy1_9'
	>>> add_impl('python2_7 pypy{1_9,2_0}', 'python3_3')
	'python{2_7,3_3} pypy{1_9,2_0}'
	>>> add_impl('python2_7', 'python3_3')
	'python{2_7,3_3}'
	>>> add_impl('python{2_{5,6},3_{1,2}} pypy{1_9,2_0}', 'python3_3')
	'python{2_{5,6},3_{1,2,3}} pypy{1_9,2_0}'
	>>> add_impl('python{2_{5,6},3_{1,2}} pypy{1_9,2_0}', 'python2_7')
	'python{2_{5,6,7},3_{1,2}} pypy{1_9,2_0}'
	>>> add_impl('python{2_{5,6},3_{1,2}} pypy{1_9,2_0}', 'pypy1_8')
	'python{2_{5,6},3_{1,2}} pypy{1_8,1_9,2_0}'
	>>> add_impl('python{2_{5,6},3_{1,2}} pypy{1_9,2_0}', 'python4_0')
	'python{2_{5,6},3_{1,2},4_0} pypy{1_9,2_0}'
	"""
	pc = parse(s)
	pc.add(new)
	return str(pc)


def del_impl(s, old):
	"""
	>>> del_impl('python2_6 python2_7 python3_2 pypy1_9', 'python2_6')
	'python2_7 python3_2 pypy1_9'
	>>> del_impl('python2_{6,7} python3_{1,2}', 'python2_6')
	'python2_7 python3_{1,2}'
	>>> del_impl('python2_{6,7} python3_{1,2,3}', 'python3_1')
	'python2_{6,7} python3_{2,3}'
	>>> del_impl('python{2_6,2_7,3_2} pypy1_9', 'python2_6')
	'python{2_7,3_2} pypy1_9'
	>>> del_impl('python{2_6,2_7} pypy1_9', 'python2_6')
	'python2_7 pypy1_9'
	>>> del_impl(' python2_6 python2_7 ', 'python2_6')
	' python2_7 '
	>>> del_impl(' python2_6 python2_7 ', 'python2_7')
	' python2_6 '
	>>> del_impl(' python2_6 python2_7 python3_2 ', 'python2_7')
	' python2_6 python3_2 '
	>>> del_impl(' python{2_{5,6,7},3_{1,2,3}} pypy{1_{8,9},2_0} ', 'python2_5')
	' python{2_{6,7},3_{1,2,3}} pypy{1_{8,9},2_0} '
	>>> del_impl(' python{2_{5,6,7},3_{1,2,3}} pypy{1_{8,9},2_0} ', 'pypy1_8')
	' python{2_{5,6,7},3_{1,2,3}} pypy{1_9,2_0} '
	"""
	pc = parse(s)
	pc.remove(old)
	return str(pc)


python_compat_re = re.compile(r'(?<![^\n])PYTHON_COMPAT=\((?P<value>.*)\)')

class EbuildMangler(object):
	def __init__(self, path):
		with open(path, 'rb') as f:
			data = f.read().decode('utf8')

		m = python_compat_re.search(data)
		if m:
			self._path = path
			self._data = data
			self._value = parse(m.group('value'))
			self._start = m.start()
			self._end = m.end()
		else:
			raise KeyError('Unable to find PYTHON_COMPAT in %s' % path)

	def write(self):
		data = ''.join((self._data[:self._start],
			'PYTHON_COMPAT=(', str(self._value), ')',
			self._data[self._end:]))

		with tempfile.NamedTemporaryFile('wb',
				dir=os.path.dirname(self._path), delete=False) as f:
			tmp_path = f.name
			f.write(data.encode('utf8'))

		os.rename(tmp_path, self._path)

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		if exc_type is None:
			self.write()

	def add(self, impl):
		self._value.add(impl)

	def remove(self, impl):
		self._value.remove(impl)
